'use strict';

const _ = require('lodash');
const path = require('path');
const yaml = require('js-yaml');
const yamlinc = require('yaml-include');
const fs = require('fs-extra');
const merge = require('lodash.merge');
const Mustache = require('mustache');
const dotenv = require('dotenv');
const execSync = require('child_process').execSync;

const getProfile = profile => (profile !== 'default' ? `--profile ${profile}` : '');

/**
 * Executes shell commands synchronously and logs the
 * stdout to console.
 * @param  {String} cmd  Bash command
 * @return {String}     The command's stdout
 */
function exec(cmd, verbose) {
  verbose = verbose || true;

  const stdout = execSync(cmd);
  if (verbose) {
    console.log(stdout.toString());
  }
  return stdout;
}

/**
 * Generates configuration arrays for ApiGateway portion of
 * the CloudFormation
 * @param  {Object} config The configuration object
 * @return {Object}        Returns ApiGateway updated configruation
 */
const configureApiGateway = (config) => {
  if (config.apis) {
    // APIGateway name used in AWS APIGateway Definition
    const apiMethods = [];
    const apiMethodsOptions = {};
    const apiDependencies = {};

    config.apis.forEach((api) => {
      apiDependencies[api.name] = [];
    });

    // The array containing all the info
    // needed to define each APIGateway resource
    const apiResources = {};

    // We loop through all the lambdas in config.yml
    // To construct the API resources and methods
    for (const lambda of config.lambdas) {
      // We only care about lambdas that have apigateway config
      if (lambda.hasOwnProperty('apiGateway')) {
        //loop the apiGateway definition
        for (const api of lambda.apiGateway) {
          // Because each segment of the URL path gets its own
          // resource and paths with the same segment shares that resource
          // we start by dividing the path segments into an array.
          // For example. /foo, /foo/bar and /foo/column create 3 resources:
          // 1. FooResource 2.FooBarResource 3.FooColumnResource
          // where FooBar and FooColumn are dependents of Foo
          const segments = api.path.split('/');

          // this array is used to keep track of names
          // within a given array of segments
          const segmentNames = [];

          segments.forEach((segment, index) => {
            let name = segment;
            let parents = [];

            // when a segment includes a variable, e.g. {short_name}
            // we remove the curly braces and underscores and add Var to the name
            if (_.startsWith(segment, '{')) {
              name = `${_.replace(_.trim(segment, '{}'), '_', '')}Var`;
            }

            name = _.upperFirst(name);
            segmentNames.push(name);

            // the first segment is always have rootresourceid as parent
            if (index === 0) {
              parents = [
                'Fn::GetAtt:',
                `- ${api.api}RestApi`,
                '- RootResourceId'
              ];
            }
            else {
              // This logic finds the parents of other segments
              parents = [
                `Ref: ApiGateWayResource${_.join(
                  _.slice(segmentNames, 0, index
                ), '')}`
              ];

              name = _.join(segmentNames.map(x => x), '');
            }

            // We use an object here to catch duplicate resources
            // This ensures if to paths shares a segment, they also
            // share a parent
            apiResources[name] = {
              name: `ApiGateWayResource${name}`,
              pathPart: segment,
              parents: parents,
              api: api.api
            };
          });

          const method = _.capitalize(api.method);
          const name = _.join(segmentNames.map(x => x), '');

          const methodName = `ApiGatewayMethod${name}${_.capitalize(method)}`;

          // Build the ApiMethod array
          apiMethods.push({
            name: methodName,
            method: _.upperCase(method),
            cors: api.cors || false,
            resource: `ApiGateWayResource${name}`,
            lambda: lambda.name,
            api: api.api
          });

          // populate api dependency list
          try {
            apiDependencies[api.api].push({
              name: methodName
            });
          }
          catch (e) {
            console.error(`${api.api} is not defined`);
            throw e;
          }

          // Build the ApiMethod Options array. Only needed for resources
          // with cors set to true
          if (api.cors) {
            apiMethodsOptions[name] = {
              name: `ApiGatewayMethod${name}Options`,
              resource: `ApiGateWayResource${name}`,
              api: api.api
            };
          }
        }
      }
    }

    return {
      apiMethods,
      apiResources: _.values(apiResources),
      apiMethodsOptions: _.values(apiMethodsOptions),
      apiDependencies: Object.keys(apiDependencies).map(k => ({
        name: k,
        methods: apiDependencies[k],
        stage: config.stage
      }))
    };
  }

  return {};
};

/**
 * Generates an array of configuration settings for
 * Lambda function in CloudFormation Template
 * @param  {Object} config The configuration object
 * @return {Object}        Returns lambdas updated configruation
 */
const configureLambda = (config) => {
  if (config.lambdas) {
    // Add default memory and timeout to all lambdas
    for (const lambda of config.lambdas) {
      if (!_.has(lambda, 'memory')) {
        lambda.memory = 1024;
      }

      if (!_.has(lambda, 'timeout')) {
        lambda.timeout = 300;
      }

      // add lambda name to services if any
      if (lambda.hasOwnProperty('services')) {
        for (const service of lambda.services) {
          service.lambdaName = lambda.name;
        }
      }

      // Get Lambda's zip file name
      lambda.zipFile = _.split(lambda.handler, '.')[0];

      if (!_.has(lambda, 'envs')) {
        lambda.envs = [];
      }
    }
  }

  return config;
};

/**
 * Generates an array of configuration settings for
 * DynamoDB Tables in CloudFormation Template
 * @param  {Object} config The configuration object
 * @return {Object}        Returns dyanmos updated configruation
 */
const configureDynamo = (config) => {
  if (config.dynamos) {
    // Add default memory and timeout to all lambdas
    for (const tb of config.dynamos) {
      // if the hash is not the first item in the schema
      // throw error
      if (_.has(tb, 'schema')) {
        if (tb.schema[0].type !== 'HASH') {
          throw Error('The first KeySchemaElement is not a HASH key type');
        }
      }
    }
  }

  return config;
};

function parseLocalEnvVariables(config) {
  try {
    const envConfig = dotenv.parse(fs.readFileSync(path.join(process.cwd(), '.env')));
    const localEnvs = {};

    Object.keys(envConfig).forEach((k) => {
      localEnvs[k] = envConfig[k];
    });

    config.localEnvs = localEnvs;
  }
  catch (e) {
    console.log('.env was not found. skipping...');
  }
  return config;
}

/**
 * Extract and parse environment variables
 * @param  {Object} config The configuration object
 * @return {Object}        Returns dyanmos updated configruation
 */
function parseEnvVariables(config) {
  let envs = {
    StackName: config.stackName,
    Stage: config.stage
  };

  // add buckets
  envs = Object.assign({}, envs, config.buckets);

  if (config.dynamos) {
    // Match tablenames to env variables name
    for (const dynamo of config.dynamos) {
      envs[dynamo.envName] = `${config.stackName}-${config.stage}-${dynamo.name}`;
    }
  }

  if (config.sqs) {
    // add sqs queues and their full names
    for (const queue of config.sqs) {
      envs[queue.name] = `${config.stackName}-${config.stage}-${queue.name}`;
    }
  }

  if (config.lambdas) {
    // add lambda names
    for (const lambda of config.lambdas) {
      if (lambda.broadcast) {
        envs[lambda.name] = `arn:aws:lambda:\$\{AWS::Region\}:\$\{AWS::AccountId\}:function:${config.stackName}-${config.stage}-${lambda.name}`; // eslint-disable-line max-len
      }
    }
  }

  // convert the objects to list for the yml template
  const envList = [];
  Object.keys(envs).forEach((env) => {
    envList.push({ key: env, value: envs[env] });
  });

  config.envsList = envList;

  return Object.assign(config, { envs: envs });
}

function loadLocalEnvs() {
  let _dotenv;
  try {
    _dotenv = dotenv.parse(fs.readFileSync(path.join(process.cwd(), '.env')));
  }
  catch (e) {
    if (e.message.includes('ENOENT')) {
      console.log('.env file is missing');
    }
    else {
      throw e;
    }
  }

  // load all env variables to an object
  return merge(process.env, _dotenv);
}

function parseStageVariables(stage = 'dev') {
  const p = path.join(process.cwd(), '.kes/stage.yml');
  const t = fs.readFileSync(p);

  // replace env variables
  const envs = loadLocalEnvs();
  Mustache.escape = (text) => text;
  const rendered = Mustache.render(t.toString(), envs);

  // convert to object from yaml
  const stageConfig = yaml.safeLoad(rendered, { schema: yamlinc.YAML_INCLUDE_SCHEMA });

  if (stage) {
    return merge(stageConfig.default, stageConfig[stage]);
  }
  return stageConfig.default;
}

/**
 * Parses the .kes/config.yml to js Object
 * @return {Object}
 */
function parseConfig(configPath, stackName, stage) {
  stackName = stackName || null;
  stage = stage || null;

  const stageConfig = parseStageVariables(stage);
  const envs = loadLocalEnvs();

  const p = configPath || path.join(process.cwd(), '.kes/config.yml');
  const configFile = fs.readFileSync(p, 'utf8');

  const v = merge(stageConfig, envs);
  Mustache.escape = (text) => text;
  const rendered = Mustache.render(configFile.toString(), v);
  let config = yaml.safeLoad(rendered, { schema: yamlinc.YAML_INCLUDE_SCHEMA });

  if (stackName) {
    config.stackName = stackName;
  }

  if (stage) {
    config.stage = stage;
  }

  config = merge(config, stageConfig);
  config = configureDynamo(config);
  config = configureLambda(config);
  config = parseEnvVariables(config);
  config = parseLocalEnvVariables(config);
  config = Object.assign(config, configureApiGateway(config));

  return config;
}

module.exports.parseConfig = parseConfig;
module.exports.exec = exec;
module.exports.getProfile = getProfile;
